from logging import LogRecord

from ...requests import Request
from ...responses import Response
from ._formatter import Formatter

__all__ = ("SimpleFormatter",)


class SimpleFormatter(Formatter):
    def format_request(self, request: Request, record: LogRecord) -> str:
        return "-> {}".format(request.url.path)

    def format_response(self, response: Response, request: Request, record: LogRecord) -> str:
        return "<- {} {}\n".format(response.status, response.reason)



import os
from distutils.util import strtobool
from typing import Optional, Union

from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import HistoryItem, HistoryRepository, HistoryRequest, HistoryResponse
from ._mock import Mock
from ._mocked import Mocked
from ._remote_handler import RemoteHandler
from ._remote_mock import RemoteMock
from ._remote_response import RemoteResponseType
from ._system_log_filter import SystemLogFilter

_REMOTE_MOCK_URL = os.environ.get("JJ_REMOTE_MOCK_URL", "http://localhost:8080")
_REMOTE_MOCK_DISPOSABLE = os.environ.get("JJ_REMOTE_MOCK_DISPOSABLE", "True")


def mocked(matcher: Union[RequestMatcher, LogicalMatcher], response: RemoteResponseType, *,
           disposable: Optional[bool] = None,
           prefetch_history: bool = True) -> "Mocked":
    if disposable is None:
        disposable = strtobool(_REMOTE_MOCK_DISPOSABLE)
    handler = RemoteMock(_REMOTE_MOCK_URL).create_handler(matcher, response)
    return Mocked(handler, disposable=disposable, prefetch_history=prefetch_history)


__all__ = ("Mock", "mocked", "RemoteMock", "RemoteHandler", "Mocked",
           "HistoryRepository", "HistoryRequest", "HistoryResponse", "HistoryItem",
           "SystemLogFilter", "RemoteResponseType",)



from inspect import isclass
from typing import Any, List, Type
from unittest.mock import sentinel as nil

from undecorated import undecorated

from ..apps import AbstractApp
from ..handlers import HandlerFunction
from ..requests import Request
from ._matcher_function import MatcherFunction
from ._registry import Registry

__all__ = ("Resolver",)


class Resolver:
    def __init__(self, registry: Registry,
                 default_app: AbstractApp,
                 default_handler: HandlerFunction) -> None:
        self._registry = registry
        self._default_app = default_app
        self._default_handler = default_handler

    def unwrap(self, fn: HandlerFunction) -> HandlerFunction:
        try:
            unwrapped = undecorated(fn)
        except ValueError:
            return fn
        return fn if unwrapped is None else unwrapped  # type: ignore

    # Apps

    def register_app(self, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.add(None, "apps", app)

    def deregister_app(self, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.remove(None, "apps", app)

    def get_apps(self) -> List[Type[AbstractApp]]:
        apps = self._registry.get(None, "apps")
        return list(apps.keys())

    # Handlers

    def register_handler(self, handler: HandlerFunction, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self.deregister_handler(handler, type(self._default_app))
        self.register_app(app)
        self._registry.add(app, "handlers", handler)

    def deregister_handler(self, handler: HandlerFunction, app: Type[AbstractApp]) -> None:
        assert isclass(app)
        self._registry.remove(app, "handlers", handler)

    def get_handlers(self, app: Type[AbstractApp]) -> List[HandlerFunction]:
        assert isclass(app)
        handlers = self._registry.get(app, "handlers")
        return list(handlers.keys())

    # Matchers

    def register_matcher(self, matcher: MatcherFunction, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self.register_handler(unwrapped, type(self._default_app))
        self._registry.add(unwrapped, "matchers", matcher)

    def deregister_matcher(self, matcher: MatcherFunction, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.remove(unwrapped, "matchers", matcher)

    def get_matchers(self, handler: HandlerFunction) -> List[MatcherFunction]:
        unwrapped = self.unwrap(handler)
        matchers = self._registry.get(unwrapped, "matchers")
        return list(matchers.keys())

    # Attributes

    def register_attribute(self, name: Any, value: Any, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.add(unwrapped, "attributes", name, value)

    def deregister_attribute(self, attribute_name: Any, handler: HandlerFunction) -> None:
        unwrapped = self.unwrap(handler)
        self._registry.remove(unwrapped, "attributes", attribute_name)

    def get_attribute(self, attribute_name: Any,
                      handler: HandlerFunction,
                      default: Any = nil) -> Any:
        unwrapped = self.unwrap(handler)
        attributes = self._registry.get(unwrapped, "attributes")
        return attributes.get(attribute_name, default)

    # Resolve

    async def _match_request(self, request: Request, matchers: List[MatcherFunction]) -> bool:
        if len(matchers) == 0:
            return False
        for matcher in matchers:
            if not await matcher(request):
                return False
        return True

    async def resolve(self, request: Request, app: AbstractApp) -> HandlerFunction:
        assert not isclass(app)
        handlers = self.get_handlers(type(app))
        for handler in reversed(handlers):
            matchers = self.get_matchers(handler)
            if await self._match_request(request, matchers):
                return handler
        return self._default_handler



from typing import List, Union, cast

from aiohttp import ClientSession
from packed import pack, unpack

from jj.http.codes import OK
from jj.matchers import LogicalMatcher, RequestMatcher

from ._history import HistoryItem
from ._remote_handler import RemoteHandler
from ._remote_response import RemoteResponseType

__all__ = ("RemoteMock",)


class RemoteMock:
    def __init__(self, url: str) -> None:
        self._url = url

    def create_handler(self,
                       matcher: Union[RequestMatcher, LogicalMatcher],
                       response: RemoteResponseType) -> RemoteHandler:
        return RemoteHandler(self, matcher, response)

    async def register(self, handler: RemoteHandler) -> "RemoteMock":
        headers = {"x-jj-remote-mock": ""}
        payload = {
            "id": str(handler.id),
            "request": handler.matcher,
            "response": handler.response,
        }
        binary = pack(payload)

        async with ClientSession() as session:
            async with session.post(self._url, data=binary, headers=headers) as response:
                assert response.status == OK, response
        return self

    async def deregister(self, handler: RemoteHandler) -> "RemoteMock":
        headers = {"x-jj-remote-mock": ""}
        payload = {
            "id": str(handler.id),
            "request": handler.matcher,
            "response": handler.response,
        }
        binary = pack(payload)

        async with ClientSession() as session:
            async with session.delete(self._url, data=binary, headers=headers) as response:
                assert response.status == OK, response
        return self

    async def fetch_history(self, handler: RemoteHandler) -> List[HistoryItem]:
        headers = {"x-jj-remote-mock": ""}
        payload = {
            "id": str(handler.id),
            "request": handler.matcher,
            "response": handler.response,
        }
        binary = pack(payload)

        async with ClientSession() as session:
            async with session.get(self._url, data=binary, headers=headers) as response:
                assert response.status == OK, response
                body = await response.read()
                unpacked = unpack(body)
                return cast(List[HistoryItem], unpacked)



from ._history_item import HistoryItem
from ._history_repository import HistoryRepository
from ._history_request import HistoryRequest
from ._history_response import HistoryResponse

__all__ = ("HistoryRepository", "HistoryRequest", "HistoryResponse", "HistoryItem",)
